package pp.s1184725.boppi;

import java.util.logging.Logger;
import java.util.stream.Collectors;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.tree.ParseTree;
import org.antlr.v4.runtime.tree.ParseTreeWalker;

import pp.s1184725.boppi.antlr.*;
import pp.s1184725.boppi.antlr.BoppiCoreParser.*;
import pp.s1184725.boppi.exception.*;
import pp.s1184725.boppi.type.*;

/**
 * This class performs type checking and variable assignment on a bare parse
 * tree.
 * 
 * @author Frank Wibbelink
 */
public class BoppiCoreChecker extends BoppiCoreBaseListener {
    public static void main(String[] args) {
        BoppiCoreLexer lexer = new BoppiCoreLexer(ToolChain.getCharStream(BoppiCoreChecker.class, "test/test2.core"));
        BoppiCoreParser parser = new BoppiCoreParser(new CommonTokenStream(lexer));
        ParseTree ast = parser.program();
        Logger logger = Logger.getLogger("TEST");
        Annotations annotations = checkProgram(ast, logger);
        System.out.println(ToolChain.getAnnotatedDOT(ast, annotations));
    }

	private Annotations an;
	private Logger log;

	/**
	 * Checks and annotates a program. Problems are reported to the given
	 * logger.
	 * 
	 * @param ast
	 *            the program's AST
	 * @param logger
	 *            the logger object to write warnings and errors to
	 * @return the generated annotations
	 */
	public static Annotations checkProgram(ParseTree ast, Logger logger) {
        BoppiCoreChecker checker = new BoppiCoreChecker(logger, new Annotations());
        ParseTreeWalker walker = new ParseTreeWalker();
        walker.walk(checker, ast);
		return checker.an;
    }

	protected BoppiCoreChecker(Logger logger, Annotations annotations) {
		an = annotations;
		log = logger;
	}

	/**
	 * Checks whether two types are compatible and log an error if they do not.
	 * 
	 * @param type1
	 *            the left hand or actual type
	 * @param type2
	 *            the right hand or desired type
	 * @param node
	 *            the context used for logging
	 */
	private void checkConstraint(Type type1, Type type2, ParserRuleContext node) {
		if (!type2.equals(type1))
			log.severe(getError(node, Messages.getString("BoppiChecker.0"), type1.toString(), type2.toString())); //$NON-NLS-1$
	}

	/**
	 * Returns an error message for a given parse tree node.
	 * 
	 * @param ctx
	 *            the parse tree node at which the error occurred
	 * @param message
	 *            the error message
	 * @param args
	 *            arguments for the message, see {@link String#format}
	 */
	private String getError(ParserRuleContext node, String message, Object... args) {
		int line = node.getStart().getLine();
		int column = node.getStart().getCharPositionInLine();
		return String.format(Messages.getString("BoppiChecker.1"), line, column, String.format(message, args)); //$NON-NLS-1$
	}

	/**
	 * Returns whether the current context resides within a given rule context
	 * within a set root (may be {@code null}). Effectively traverses between
	 * {@code current} and {@code root} in the parent hierarchy looking for a
	 * {@code rule} class.
	 * 
	 * @param <T>
	 *            the rule class
	 * @param ctx
	 *            the node to start traversing
	 * @param root
	 *            the node to stop traversing
	 * @param rule
	 *            the ParserRuleContext class to look for
	 * @return whether a matching rule is found between the current and root
	 *         node
	 */
	private <T extends ParserRuleContext> boolean isWithinRule(ParserRuleContext ctx, ParserRuleContext root,
			Class<T> rule) {
		for (ParserRuleContext node = ctx; node != root && node != null; node = node.getParent())
			if (node.getClass().isAssignableFrom(rule))
				return true;

		return false;
	}
}
